/*
    $Id: InputData.java,v 1.1 1997/07/23 18:00:53 augusto Exp $

    Copyright (C) 1997  Augusto Ciuffoletti

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact address: augusto@di.unipi.it

    $Log: InputData.java,v $
    Revision 1.1  1997/07/23 18:00:53  augusto
    Initial revision

*/
package syd.inputdata;

import java.util.StringTokenizer;

/**
 * A manager for the data retrieved from the command line and from the 
 * configuration file. Most fields are public, and this part of the
 * program should be revised for style. Each object represents
 * the data associated to a certain session.
 *
 * @version $Id: InputData.java,v 1.1 1997/07/23 18:00:53 augusto Exp $
 * @author Augusto Ciuffoletti
 *
 */
public class InputData {
  private DefaultValues d_v;
  private StringBuffer b;
  private StringTokenizer par_tok;

  /** the identifier of the session */
  public int identity;
  /** IP address of the server */
  public String host;
  /** upper bound of server's drift in parts per millions */
  public double rho; 
  /** lower bound of the communication delay (msecs) */
  public long min;
  /** the "contact reached" precision threshold (msecs) */
  public long epsilon; 
  /** the worst admitted precision threshold (msecs) */
  public long delta_max;
  /** delay between successive attempts (msecs) */
  public long w;       
  /** the number of attempts before diagnosing unreachability */
  public int k;
  /** the name of the log file */
  public String logfile;
  /** if TRUE, be verbose */
  public boolean verbose;
  /** LOGICAL_CLOCK, or SET_CLOCK, or ADJUST_CLOCK */
  public int mode;    
  /** the port used by the clock synch. protocol */
  public int port;     

  /**
   * CONSTRUCTOR: Initializes a new object taking the values from
   * the <code>DefaultValues</code> class.
   *
   * @param i the identifier of the session
   * @see #DefaultValues
   */
  public InputData(int i) {
    identity=i;
    d_v = new DefaultValues();
    verbose = false;
    host=d_v.HOST;
    min=d_v.MIN;
    rho=d_v.RHO;
    epsilon=d_v.EPSILON;
    delta_max=d_v.DELTA;
    w=d_v.W;
    k=d_v.K;
    logfile=d_v.HOME+"/log_"+i;
    mode=0;
    port=d_v.PORT;
    return;
  }
  /**
   * ACCESSOR: checks that the data provided are consistent, and print
   * a diagnostic on the standard error output.
   */
  public void check_consistency() throws DataConsistencyException {
    if ( (delta_max-epsilon)*1000000/rho < k * w ) 
      { System.err.println("Input data for session " +identity+ " are inconsistent!");
      System.err.println("The time needed to perform k attempts is greater than the time");
      System.err.println("to drift from the precision epsilon to the upper bound delta_0.");
      System.err.println("The following must hold:");
      System.err.println("(delta_max-epsilon)*1000000/rho < k * w");
      DataConsistencyException e = new DataConsistencyException();
      e.where = "nil";
      e.why = "";
      throw e;
    }
  }
  /**
   * MUTATOR: parses the command line and updates this InputData.
   *
   * @param args a String[] in the form in which command line parameters
   * are passed to a <code>main</code>
   * @throws CommandLineException if a parse error occurs
   */
  public void parse_command_line(String[] args) throws CommandLineException {
    int i;
    for (i=0; i<args.length;i++) {
      if (args[i].charAt(0) != '-') {
        CommandLineException e = new CommandLineException();
        e.why = "Bad command line. Option expected, got ";
        e.what = args[i];
        throw e;
      }
      if (args[i].length() != 2) {
        CommandLineException e = new CommandLineException();
        e.why = "Bad Command Line. Only single character options: ";
        e.what = args[i];
        throw e;
      }
      switch (args[i].charAt(1)) {
      case 'h':
        i++;
        host = args[i];
        break;
      case 'm':
        i++;
        min = Integer.valueOf(args[i]).intValue();
        break;
      case 'r':
        i++;
        rho = Integer.valueOf(args[i]).intValue();
        break;
      case 'e':
        i++;
        epsilon = Integer.valueOf(args[i]).intValue();
        break;
      case 'd':
        i++;
        delta_max = Integer.valueOf(args[i]).intValue();
        break;
      case 'w':
        i++;
        w = Integer.valueOf(args[i]).intValue();
        break;
      case 'k':
        i++;
        k = Integer.valueOf(args[i]).intValue();
        break;
      case 'v':
        verbose = true;
        break;
      case 's':
        mode = 0;
        break;
      case 'a':
        mode = 0;
        break;
      default:
        CommandLineException e = new CommandLineException();
        e.why = "Bad Command Line. Undefined option: ";
        e.what = args[i];
        throw e;
      }
    }
  }    
  /**
   * MUTATOR: parses a string and updates this InputData.
   *
   * @param args a String in the form in which it is encoded in the
   * configuration file.
   * @throws ConfigurationLineException if a parse error occurs
   */

  public void parse_String(String par) throws ConfigurationLineException {
    String tmp;

    par_tok = new StringTokenizer(par);
    if (par_tok.countTokens() != 6) {
      ConfigurationLineException e = new ConfigurationLineException();
      e.why = "Bad Configuration line (6 items needed; default is -)";
      e.what = par;
      throw e; 
    }
    if ( par_tok.hasMoreTokens() ) {
      tmp = par_tok.nextToken();
      if ( !tmp.equals("-") )
        host = tmp;
    }
    if ( par_tok.hasMoreTokens() ) {
      tmp = par_tok.nextToken();
      if ( !tmp.equals("-") )
        min = Integer.valueOf(tmp).intValue();
    }
    if ( par_tok.hasMoreTokens() ) {
      tmp = par_tok.nextToken();
      if ( !tmp.equals("-") )
        epsilon = Integer.valueOf(tmp).intValue();
    }
    if ( par_tok.hasMoreTokens() ) {
      tmp = par_tok.nextToken();
      if ( !tmp.equals("-") )
        delta_max = Integer.valueOf(tmp).intValue();
    }
    if ( par_tok.hasMoreTokens() ) {
      tmp = par_tok.nextToken();
      if ( !tmp.equals("-") )
        w = Integer.valueOf(tmp).intValue();
    }
    if ( par_tok.hasMoreTokens() ) {
      tmp = par_tok.nextToken();
      if ( !tmp.equals("-") )
        k = Integer.valueOf(tmp).intValue();
    }
    return;
  }
  /**
   * ACCESSOR: builds a String that contains all relevant in[ut data
   * that are specific to the session this object refers to
   *
   * @return a String containing a printable report
   */
  public String toString() {
    b = new StringBuffer();
    b.append(identity+"| ======== Synchronization session header ========\n");
    b.append(identity+"| NTP server " + host);
    b.append("\t logfile: " + logfile + "\n");
    b.append(identity+"| min = " + min + " msecs");
    b.append("\tepsilon = " + epsilon + " msecs");
    b.append("\tdelta_max = " + delta_max + " msecs\n");
    b.append(identity+"| W = " + w + " secs");
    b.append("\t\tk = " + k + " times\n");
    return b.toString();
  }
  /**
   * ACCESSOR: builds a String that contains all relevant input data
   * that are common to all sessions
   *
   * @return a String containing a printable report
   */
  public String common_header() {
    b = new StringBuffer();
    b.append("======== Common header ========\n");
    b.append("Mode: ");
    if (verbose)
      b.append("verbose ");
    else
      b.append("quiet ");
    switch (mode) {
    case 0:
      b.append("logical clock;");
      break;
    case 1:
      b.append("set clock;");
      break;
    case 2:
      b.append("adjust clock;");
      break;
    }
    b.append("\trho = " + rho + " ppm;\n");
    return b.toString();
  }
}



