/*
    $Id: LogicalClock.java,v 1.1 1997/07/23 18:36:39 augusto Exp $

    Copyright (C) 1997  Augusto Ciuffoletti

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact address: augusto@di.unipi.it

    $Log: LogicalClock.java,v $
    Revision 1.1  1997/07/23 18:36:39  augusto
    Initial revision

*/
package syd.logicalclock;

import syd.remoteclock.NTPTime;
import java.util.*;

/**
 * Real times are all expressed as long integer values, one unit corresponding
 * to one millisecond, according with java APIs that allow a granularity of 1 
 * millisecond.
 * <p>
 * IMPORTANT: this object is built to be "multiple readers, single writer".
 * The only Accessor that can be reliably operated by several
 * partners is the <code>toBytes</code>
 * and is presently the only
 * <code>synchronized</code> method. The chances that its operation slows
 * the other operations is considered negligible: otherwise, the "origin" 
 * field could be used as a timestamp to abort this operation without 
 * requiring a the <code>synchronized</code> operation: checking the value
 * before of this fiels before and after the construction of the
 * <code>byte</code> array one understands if other values were updated.
 * The <code>toString</code> is non reentrant.
 * @see #toString
 * @see #toBytes
 *
 * @version $Id: LogicalClock.java,v 1.1 1997/07/23 18:36:39 augusto Exp $
 * @author Augusto Ciuffoletti
 *
 */

public class LogicalClock{
  private boolean is_valid = false;  /* if true the clock precision is bound */
  private boolean commit = false;    /* if true the clock value can be used */
  private long origin;               /* the time when the clock was set up */
  private long precision;            /* the upper bound of the difference
                                        between the log_clock and the time
					reference (e.g. UTC) when the clock
					was set up */
  private long offset;               /* the offset between the system clock
			                and the log_clock when the log_clock 
				        was set */
  private double drift = 0;            /* the speed with which the log_clock
                                        runs apart from the system clock, in
					usecs per second (i.e. parts per
					million) (unused) */
  private double rho;	             /* the upper bound of the drift of the
                                        system clock with respect to the
					reference time, in parts per million */
  /** 
   * CONSTRUCTOR
   *
   * @param r is the maximum drift (rho) in [arts per million
   */
  public LogicalClock(double r){
    origin=SystemClock.read();
    rho=r;
  }
  /** 
   * ACCESSOR: returns the offset of the logical clock with respect to
   * system clock.
   *
   * @return the difference from the logical clock and the system clock
   */
  public long offset() {
    return offset;
  }
  /** 
   * ACCESSOR: tests whether this clock can be used or not 
   *
   * @return a boolean that is true if the clock is valid
   */
  public boolean is_valid() {
    return is_valid;
  }
  /** 
   * MUTATOR: commits this clock as accurate 
   *
   */
  public void commit() {
    commit=true;
  }
  /** 
   * ACCESSOR: tests whether this clock is accurate or not. 
   *
   * @return a boolean that is true if the clock is within the tolerance
   */
  public boolean is_committed() {
    return commit;
  }
  /**
   * These are the basic methods to set and read the logical clock. There are 
   * also displaced methods that can be used to read and set a logical clock 
   * value at a given a baseline
   */
  /** 
   * ACCESSOR: Returns the current value of this logical clock.
   *
   * @return a long that represents the logical clock in millisecs (Java)
   * @throws LogicalClockException if the clock is not valid
   */  
  public long value() 
                     throws LogicalClockException {
    long t;
    LogicalClockException e;
    if ( is_valid ) {
      return SystemClock.read()+offset;
    } else
      throw new LogicalClockException();
  }
  /** 
   * ACCESSOR: Returns the current precision of this logical clock.
   *
   * @return a long that represents the estimated error (in millisecs)
   * @throws LogicalClockException if the clock is not valid
   */  
  public long precision() 
                     throws LogicalClockException {
    long t;
    LogicalClockException e;
    if ( is_valid ) {
      return  (long) ( precision + 
                       ( (SystemClock.read() - origin) * rho / 1000000 ) );
    } else
      throw new LogicalClockException();
  }
  /** 
   * MUTATOR: Sets the current value and precision of this logical clock.
   *
   * @param v the current time in millisecs (Java)
   * @param p the estimated precision in millisecs
   */  
  public void set(long v, long p) {
    origin = SystemClock.read();
    offset = v - origin;
    precision = p;
    is_valid = true;
    return;
  }
  /** 
   * These are the additional methods to set and read the logical clock at a 
   * time different from the current time. This time is called "baseline",
   * and corresponds to a time measured on the system clock
   **/
  /** 
   * ACCESSOR: Returns the value of this logical clock at a given time.
   *
   * @param when the baseline (computed on the system clock) 
   * @return a long that represents the logical clock in millisecs (Java)
   * @throws LogicalClockException if the clock is not valid
   */  
  public long value(long when) 
                     throws LogicalClockException {
    long t;
    LogicalClockException e;
    if ( is_valid ) {
      return when+offset;
    } else
      throw new LogicalClockException();
  }
  /** 
   * ACCESSOR: Returns the precision of this logical clock at a given time.
   *
   * @param when the baseline (computed on the system clock) 
   * @return a long that represents the estimated error (in millisecs)
   * @throws LogicalClockException if the clock is not valid
   */  
  public long precision(long when) 
                     throws LogicalClockException {
    long t;
    LogicalClockException e;
    if ( is_valid ) {
      return  (long) ( precision + 
                       ( (when - origin) * rho / 1000000 ) );
    } else
      throw new LogicalClockException();
  }
  /** 
   * MUTATOR: Sets the value and precision of this logical clock 
   * at a given time.
   *
   * @param v the time in millisecs (Java)
   * @param p the estimated precision in millisecs
   * @param when the baseline (computed on the system clock) 
   */  
  public synchronized void set(long v, long p, long when) {
    origin = when;
    offset = v - origin;
    precision = p;
    is_valid = true;
  }
  /** 
   * The following methods are used to transmit the "logical time" through
   * a socket: the relevant part of the state is encoded into and decoded from
   * a byte array.
   */
  /** 
   * ACCESSOR: encodes the relevant variables into a new byte array.
   *
   * @return a new byte array
   */
  public synchronized byte[] toBytes() {
    int i;

    byte[] packet = new byte[40];
    byte[][] format = new byte[5][8];
    format[0] = encode(origin);
    format[1] = encode(precision);
    format[2] = encode(offset);
    format[3] = encode((long)(drift * 1000));
    format[4] = encode((long)(rho * 1000));
    for (i=0;i<40;i++) packet[i]=format[i/8][i%8];
    return packet;
  }
  /**
   * CONSTRUCTOR: creates a new logical clock using an array of bytes to
   * determine the relevant fields.
   *
   * @param packet the byte array that encodes the internal state;
   * @return a new <code>LogicalClock</code>
   **/ 
  public LogicalClock(byte[] packet) {
    int i;

    byte[][] format = new byte[5][8];
    for (i=0;i<40;i++) format[i/8][i%8]=packet[i];
    origin = decode(format[0]);
    precision = decode(format[1]);
    offset =  decode(format[2]);
    drift =  (double)decode(format[3])/1000;
    rho = (double)decode(format[4])/1000;
    is_valid=true;
    commit=true;
  }
  /**
   * ACCESSOR: this is mainly for testing purpose
   */
  public String toString() {
    StringBuffer s = new StringBuffer();
    s.append("Validity is "+is_valid+"\n");
    s.append("Commitment is "+commit+"\n");
    s.append("Origin is: "  + new Date(origin) + " " + (origin % 1000) + " msecs\n");
    s.append("Precision (at origin) is "+precision+" msecs\n");
    s.append("Offset is "+offset+" msecs\n");
    s.append("Drift is "+drift+" ppm\n");
    s.append("Rho is "+rho+" ppm\n");
    return s.toString();
  }
  /* The following are two helpers that make more readable the encoding
     and decoding methods on the whole object */
  private byte[] encode(long x) {
    byte[] data = new byte[8];
    for (int j=7;j>=0;j--) {
      data[j] = (byte) (x & 255);
      x = x >>> 8;
    }
    return data;
  }
  private long decode(byte[] data) {
    int j;
    long y;
    long result=0;

    for (j=0;j<8;j++) {
      y = data[j] & 255;
      result = (result << 8) | y;
    }
    return result;
  }
}










