/*
    $Id: Mpx.java,v 1.1 1997/07/23 18:37:08 augusto Exp $

    Copyright (C) 1997  Augusto Ciuffoletti

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact address: augusto@di.unipi.it

    $Log: Mpx.java,v $
    Revision 1.1  1997/07/23 18:37:08  augusto
    Initial revision

*/
package syd.mpx;

/**
 * The multiplexer Mpx is a basic a-symmetric synchronization tool. It
 * implements a non-deterministic request delivery mechanism. Several entities
 * (emitters) can submit requests to another entity (collector). Each emitter
 * is distinguished by a unique key. The requests are delivered one at a time,
 * and the emitter waits until its request is delivered. The collector waits
 * untils a new request is submitted, if none is present. The collector is also
 * informed when there are no more alive emitters.
 * <p>
 * Intended use:
 * <ul>
 * <li> 
 * Each emitter is given a different identifier, and sets itself as alive
 * calling the "connect" method of the Mpx. The collector does not execute
 * any "select" statement before calling the "accept" Mpx method.
 * <li>
 * The collector loops around a "select" statement serving the requests, until
 * the "select" finally returns a negative value. On the other side, the
 * emitter loops around a "submit" statement. Once the emitter wants to
 * terminate the session, it calls the "quit" method, to cancel itself from
 * the set of "alive" emitters.
 * <li>  
 * When the emitter submita a request to the collector, it blocks until the
 * request is delivered: therefore it could block forever if the collector
 * does not execute any more "select" methods.
 * </ul>
 * <p>
 * The Mpx is primarily intended as multiplexer among several warm spare
 * activities (the emitters): each of them periodically reports about its
 * activity to one centralized entity (the collector), which coordinates
 * their activity using such reports.
 * <p>
 * The responsability for its correct use is primarily of the program: the 
 * Mpx does not enforce the above policy. 
 *
 * @version $Id: Mpx.java,v 1.1 1997/07/23 18:37:08 augusto Exp $
 * @author Augusto Ciuffoletti
 *
 */

public class Mpx {
  EmitterStates state;
  static int next;
  static int how_many;

  /**
   * CONSTRUCTOR: creates a new multiplexer. No request pending in this 
   * state, and all emitters are considered not alive.
   */
  public Mpx(int n) {
    how_many = n;
    state = new EmitterStates(n);
    next = 0;
    return;
  }

  /** 
   * ACCESSOR: returns the identifier if the first, with respect to the
   * ordering of the identifiers, emitter that is alive.
   *
   * @return an int indicating the emitter, or <code>-1</code> if
   * there are no alive emitters 
   */
  public synchronized int first_alive() {
    int i;
    for (i=0; i<how_many; i++)  
      if (state.is_alive(i)) return i;
    return -1;
  }

  /** 
   * ACCESSOR: returns the identifier if the next, with respect to the
   * ordering of the identifiers, emitter that is not alive. 
   * <p>
   * The Mpx is remindful of the last answer to this method, that therefore 
   * implements a fair, although predictable, choice among non not alive
   * emitters.
   *
   * @return an int indicating the emitter, or <code>-1</code> if
   * there are no alive emitters 
   */
  public synchronized int next_not_alive() {
    int i;
    for (i=0; i<how_many; i++) {
      next=(next+1) % how_many; 
      if (!state.is_alive(next)) 
	return next;
    }
    return -1;
  }

  /**
   * ACCESSOR: tests whether the given emitter is alive or not.
   * Collector side.
   *
   * @param key the identifier of an emitter
   * @returns a <code>boolean</code>, <code>true</code> if the emitter
   * is alive
   */
  public synchronized boolean is_alive(int key) {
    return state.is_alive(key);
  }

  /**
   * ACCESSOR: blocks until the given emitter becomes alive. Collector side
   *
   * @param key the identifier of an emitter
   * @throws InterruptedException if the <code>wait</code> is interrupted
   */
  public synchronized void mpx_accept(int key) throws InterruptedException {
    while ( ! state.is_alive(key) ) {
      wait();
     }
    return;
  }

  /**
   * ACCESSOR: blocks until a request is submitted, or there are no more 
   * alive emitters. Collector side.
   * <p>
   * In the first case the key of the requesting emitter is returned, 
   * in the latter the return value is negative.
   *
   * @return an int indicating the emitter that submitted a request, or
   * that there are no alive emitters 
   * @throws InterruptedException if the <code>wait</code> is interrupted
   */
  public synchronized int mpx_select() throws InterruptedException {
    int i;
    while (state.no_pending()) {
      if ( state.no_alive() ) return -1;
        wait();
      if ( state.no_alive() ) return -1;
    }
    i=state.next_pending();
    state.reset_pending(i);
    notifyAll();
    return i;
  }

  /**
   * MUTATOR: sets the emitter as alive. Emitter side.
   *
   * @param key an int that indicates an emitter
   * @throws EmitterAliveException if the emitter is already alive.
   */
  public synchronized void mpx_connect(int key) throws EmitterAliveException {
    if (is_alive(key)) {
      EmitterAliveException t = new EmitterAliveException();
      t.key=key;
      throw t;
    }
    state.set_alive(key);
    notifyAll();
  }

  /**
   * MUTATOR: submits a request and wait for the request to be delivered;
   * Emitter side.
   *
   * @param key an int that indicates an emitter
   * @throws InterruptedException if the <code>wait</code> is interrupted
   */
  public synchronized void mpx_submit(int key) throws InterruptedException {
    while (state.is_pending(key)) {
      wait();
    }
    state.set_pending(key);
    notifyAll();
    return;
  }

  /**
   * MUTATOR: declares itself as not alive; emitter side.
   *
   * @param key an int that indicates an emitter
   * @throws InterruptedException if the <code>wait</code> is interrupted
   */
  public synchronized void mpx_quit(int key) throws InterruptedException {
    while (state.is_pending(key)) {
      wait();
    }
    state.reset_alive(key);
    notifyAll();
    return;
  }
}

