/*
    $Id: RemoteClock.java,v 1.2 1997/07/25 17:54:39 augusto Exp $

    Copyright (C) 1997  Augusto Ciuffoletti

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact address: augusto@di.unipi.it

    $Log: RemoteClock.java,v $
    Revision 1.2  1997/07/25 17:54:39  augusto
    Eliminato il riferimento a InputData nei parametri

    Revision 1.1  1997/07/23 18:37:39  augusto
    Initial revision

*/
package syd.remoteclock;

import java.net.*;
import java.io.*;
import java.util.*;
import java.lang.*;
import syd.inputdata.*;
import syd.logicalclock.*;

/**
 * Gets in touch with a remote NTP server and holds that data about the
 * contact.
 *
 * For details about the NTP protocol please see
 * <A HREF="http://sunsite.auc.dk/RFC/rfc/rfc1305.html"> rfc 1305 </A>
 *
 * @version $Id: RemoteClock.java,v 1.2 1997/07/25 17:54:39 augusto Exp $
 * @author Augusto Ciuffoletti
 *
 */
public class RemoteClock {
  private long retry_baseline;  // the baseline of the current retry
  private long read_baseline;   // the baseline of the reading operation
  private long t0,tr,tx,t1;     // timing of the remote clock reading
  private long precision;       // server's precision
  private String host;          // server's address
  private int port;             // server's port (NTP is 123)

  /**
   * MUTATOR: carries out the communication with the remote NTP server.
   * <p>
   * The operation is timed out using the quantity <code>w</code>
   * passed as parameter, computed from a baseline <code>b</code> passed 
   * as parameter. The timing of the operation is computed using a 
   * <code>LogicalClock</code>, also provided as a parameter. If the
   * <code>LogicalClock</code> is not valid, the <code>SystemClock</code>
   * is used instead.
   *
   * @param b a long indicating the baseline upon which to count the
   * timeout;
   * @param lc the <code>LogicalClock</code> used to compute the
   * timing of the operation.
   * @param h a String containing the address (IP or symbolic) of the
   * server
   * @param p the port number at the server's end: 123 for NTP
   * @param w a long indicating the lapse indicating the timeout
   * @throws IOException if a network error occurs;
   * @throws LogicalClockException if there is a problem using the 
   * logical clock;
   */
  public void read(long b, LogicalClock lc, String h, int p, long w) 
         throws IOException, LogicalClockException {
    long stamp;     // timestamp for message identification
    long to;        // the received timestamp
    long timeout;   // the timeout for the receive operation
    long fuzz;      // additional error added by the baseline recording
    boolean flag;   // true if the logical clock provided is valid
    NTPTime t0_ntp, to_ntp, tr_ntp, tx_ntp;
    int i,j,k;

    host=h;
    port=p;
    byte[] send_buffer = new byte[48];
    InetAddress address;
    DatagramPacket send_packet;
    DatagramPacket receive_packet;
    DatagramSocket socket;
    byte[][] a;
/* data to be included in the message */
    int pac_LI=0;   /* 0=no leap second */
    int pac_VN=3;   /* NTP Version number (3) */
    int pac_Mode=3; /* Operation mode. 3=client */
    int pac_Stratum=3; /* Stratum. 255=minimum */

    retry_baseline=b;
    t0=0;
    t1=0;
    a=new byte[6][8];
    socket = new DatagramSocket();
    address = InetAddress.getByName(host);
    a[0][0] = (byte) ( (pac_LI * 64) + (pac_VN * 8) + pac_Mode );
    flag = lc.is_valid();
    if (flag)
      stamp = lc.value();
    else
      stamp = SystemClock.read();
    a[5] = (new NTPTime(stamp)).toBytes();  /* assigning transmit time */
/* loading data into packet */
    for (i=0;i<48;i++) send_buffer[i]=a[i/8][i%8];
    send_packet=new DatagramPacket(send_buffer,48,address,port);
// ==========================================
// The following is a time critical section: the time taken to complete it
// determines the clock reading precision.
    if (flag)
      t0 = lc.value();
    else
      t0 = SystemClock.read();
    socket.send(send_packet); // sending packet
    do {
      timeout= w - ( SystemClock.read() - retry_baseline );
      if ( timeout < 0 )
        throw new IOException("Timed out while receiving wrong messages");
      socket.setSoTimeout((int)timeout); // set the timeout
      receive_packet=new DatagramPacket(send_buffer,48);
      socket.receive(receive_packet); // waiting for response
      read_baseline=SystemClock.read();
      if (flag)
        t1 =  lc.value();
      else
        t1 = SystemClock.read();
      fuzz = read_baseline-SystemClock.read();
      fuzz = Math.abs(fuzz);
// End of time critical section.
// ==========================================
// the packet is decomposed into parts of 8 bytes
      for (i=0;i<48;i++) a[i/8][i%8]=send_buffer[i];
// the parts containing times are decoded from ntp representation
      to = new NTPTime(a[3]).toLong();
      tr = new NTPTime(a[4]).toLong();
      tx = new NTPTime(a[5]).toLong();
    } while (to != stamp);
    socket.close();
    precision=0;
    for (i=0;i<4;i++) {
      j=(int) a[1][i];
      if (j<0) j += 256;
      precision = (precision * 256) + j;
    }
    precision=(long) (precision * 1000 / 65536);
    /*    precision=(long)(Math.pow(2,a[0][3])*1000)+fuzz; */
    return;
  }
//==============================================
//             Accessor methods
//==============================================
  /**
   * ACCESSOR: returns the baseline (computed on the <code>SystemClock</code>)
   * corresponding to the event of receiving the reply from the server.
   *
   * @return the baseline (computed on the <code>SystemClock</code>)
   * corresponding to the event of receiving the reply from the server.
   */
  public long read_baseline() {
    return read_baseline;
  }
  /**
   * ACCESSOR: returns the time when the request was originated (according 
   * to the logical clock or to the system clock).
   *
   * @return the time when the request was originated (according to the
   * logical clock or to the system clock).
   * @see #read
   */
  public long t0() {
    return t0;
  }
  /**
   * ACCESSOR: returns the time when the request was received by the server 
   * (according with server's clock)
   *
   * @return the time when the request was received by the server (according
   * with server's clock)
   */
  public long tr() {
    return tr;
  }
  /**
   * ACCESSOR: returns the time when the reply was sent by the server 
   * (according with server's clock)
   *
   * @return the time when the reply was sent by the server (according
   * with server's clock)
   */
  public long tx()  {
    return tx;
  }
  /**
   * ACCESSOR: returns the time when the reply was received (according to the
   * logical clock or to the system clock).
   * 
   * @return the time when the reply was received (according to the
   * logical clock or to the system clock).
   * @see #read
   */
  public long t1()  {
    return t1;
  }
  /**
   * ACCESSOR: returns the DISPERSION (sorry about the name) declared by the 
   * server for its own clock 
   *
   * @return the DISPERSION (sorry about the name) declared by the server
   * for its own clock
   */
  public long precision()  {
    return precision;
  }
}










