/*
    $Id: Session.java,v 1.3 1998/02/12 15:53:46 augusto Exp $

    Copyright (C) 1997  Augusto Ciuffoletti

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact address: augusto@di.unipi.it

    $Log: Session.java,v $
    Revision 1.3  1998/02/12 15:53:46  augusto
    The retry method has been rewritten in order to eliminate a small
    bug. In essence, the remote dispersion was not taken into account
    in some steps. The structure of the program has been significantly
    improved...

    Revision 1.2  1997/07/25 17:55:19  augusto
    Eliminato il riferimento a InputData nei parametri alla RemoteClock

    Revision 1.1  1997/07/23 18:38:24  augusto
    Initial revision

*/
package syd.session;

import java.io.*;
import java.lang.*;
import java.util.*; /* for Date */
import java.net.*; /* for Date */
import syd.mpx.*;
import syd.inputdata.*;
import syd.logicalclock.*;
import syd.remoteclock.*;
import syd.log.*;
import syd.Daemon;

/**
 *
 * @version $Id: Session.java,v 1.3 1998/02/12 15:53:46 augusto Exp $
 * @author Augusto Ciuffoletti
 *
 */

public class Session extends Thread{
  private int identity;		  // the identifier for this session
  private InputData sd;		  // the session description 
  private Log log;	          // the logfile assigned to this session 
  private LogicalClock lc;	  // the logical clock kept by this session 
  private RemoteClock rc;	  // the remote clock as it is read 
  private boolean is_accurate;	  // the LogicalClock is valid 
  private Mpx channel;		  // the channel used with the Daemon 
  private long retry_baseline;    // the time baseline of a retry 
  int count;                      // how many retries are left 

  /**
   * CONSTRUCTOR: a new <code>Session</code> is built.
   *
   * @param p_1 the description of this session (host, min etc...)
   * @param p_2 the multiplexer used to synchronize with the daemon
   * @param p_3 the logical clock controlled by this session
   */
  public Session(
        InputData p_1, 
        Mpx p_2, 
        LogicalClock p_3) {
    sd = p_1;
    identity = sd.identity;
    channel = p_2;
    lc = p_3;
    return;
  }

  /**
   * MUTATOR: it is the basic method that is used to run the session;
   * it returns when the clock synchronization algorithm fails.
   * <p>
   * The error conditions that may arise as a consequence of an unexpected
   * behavior of the system clock (eccessive drift, or resynchronizations)
   * are coped with by forgetting the past value of delta_0, and
   * recomputing the value of the clock using the "INITIAL" rule (see paper)  
   */
  public void run() {
    String intestazione;
    StringBuffer report = new StringBuffer();
    long dna;
    
/* Creating a new remote clock */
    rc = new RemoteClock();
    try {
      log=new Log(sd);           // Creating (or re-create) the log file.
      channel.mpx_connect(identity);  // Get connected to the Multiplexer
      while ( true ) {                // The following is the main loop    
	attempt();                    // Perform the attempt
        channel.mpx_submit(identity); // and inform daemon about success 
        dna = (long) ( ( sd.delta_max - lc.precision() ) /
                       ( sd.rho / 1000000 ) ) -
                     ( sd.k * sd.w );
        log.verbose("Delay to next attempt (dna) is "+ (dna / 60000) + "\' " + (dna % 60000)/1000 + "\" ("+dna+" msecs)\n");
	if (dna < 0) dna = 0;
        Thread.sleep(dna);
      }
/* end of the main loop: now catching interruptions */
    } catch (EmitterAliveException e) {
      System.err.println("Session "+identity+" already alive");
      try {
      log.close();
      } catch ( IOException ex ) {
      }
      e.printStackTrace();
      return;
    } catch (Exception e) {
      e.printStackTrace();
      try { 
        log.verbose("Session " + identity + " aborts");
        log.close();
      } catch ( IOException ex ) {
      }
      try { 
        channel.mpx_quit(identity); 
      } catch ( InterruptedException ex ) { 
        ex.printStackTrace(); 
      }
      return;
    }
  }
  void attempt() 
       throws ProtocolFailureException,IOException,
              InterruptedException,LogicalClockException {
    long sleep_time;          // the time before next attempt
    long delta_0=0;           // the LogicalClock precision
    boolean must_sleep=false; // the retry must be performed immediately 
    StringBuffer report = new StringBuffer();

    count = sd.k;
    log.verbose("  New attempt");
    retry_baseline = 0;    // the initial value is immaterial   
    do {
/* wait for lower traffic */
      if ( must_sleep )
        Thread.sleep(sd.w - (SystemClock.read() - retry_baseline));
      retry_baseline = SystemClock.read();  /* setting retry baseline */
      must_sleep = retry();           /* retrying remote clock reading */
/* exiting the following "try", either the delta_0 has a valid value and the
   is_accurate boolean is true, or the is_accurate boolean is false */
      try {
        delta_0 = lc.precision();
        is_accurate = (delta_0 < sd.delta_max ); 
      } catch ( LogicalClockException e ) {
        is_accurate = false;
      }
/* if the clock becomes non accurate after being released as accurate,
   throw an exception */
      if ( ! is_accurate && lc.is_committed() ) {      
        throw new ProtocolFailureException("Session "+ identity + ": Contact lost!");
      }
      count--;
    } while ( 
         (count > 0) && 
         ( ! is_accurate || 
           delta_0 > sd.epsilon 
         )
      );
/* The synchronization attempt is over, either successfully or not */ 
/* TODO: log failure */ 
    if ( ! is_accurate ) 
      throw new ProtocolFailureException("Session "+ identity + ": Cannot reach contact");
    lc.commit();
    report.append("  Attempt was successful after "+(sd.k-count)+" retries\n");
    report.append("  Final result was ");
      if (delta_0 > sd.epsilon)
        report.append("best effort, ");
      else
        report.append("reached contact, ");
    report.append("with precision delta_0 = " + delta_0 + " msecs\n");
    log.verbose(report.toString());
/* TODO: log success */
  }
/* =================================================
   This function applies the reading rule, that calculates an approximation
   of UTC time based on the timing of the remote clock reading and of the 
   previous precision of the logical clock LC.  The new approximation is 
   used to update the logical clock.

   The retry method either returns false and lasts exactly "W" time units
   or returns true, and lasts not more than "W" time units from the time 
   recorded in the retry_baseline field.
   The function returns true if the attempt performed correctly, false if a 
   timeout occurred during the attempt (the deadline is W secs fromt the 
   beginning of the attempt). The method re-throws the same exceptions
   thrown by the remote clock reading.

   The code is strictly adherent to that presented in the paper, and it is
   not annotated.
   ================================================= */
  boolean retry() 
          throws ProtocolFailureException, InterruptedException, 
                 IOException, LogicalClockException {
    long alpha_max, alpha_min;  // see paper 
    long beta_max, beta_min;    // see paper 
    long psi;			// see paper 
    long t;                     // new time value
    long delta_0;        	// new precision
    long delta_sum;
    long delta_0_old=Long.MAX_VALUE;// estimated precision before the retry 
    String mode;                // the synchronizazion mode used: 
                                // NOT_SYNC: initial value;
                                // USELESS, CRISTIAN, ALA_CIU: see paper
    long l1, l2;		// threshold values (see paper) 
    long D;			// temporary value (see paper) 
    long x;			// temporary value
    boolean emergency=false;    // true when consistency is violated 
    long errdelta=0;            // the lower bound of the inconsistent delta
    StringBuffer report = new StringBuffer();

    // fills the header of the report
    log.verbose("    New retry");
    if ( ! lc.is_valid() )
      {
	delta_0 = 0;
        report.append("    Initial precision is unbound\n");
      }
    else
      {
        delta_0 = lc.precision();
        report.append("    Initial precision is: " + delta_0 + " msecs\n");
      }
    emergency = false;
    // reads the remote clock
    try {
      rc.read(retry_baseline, lc, sd.host, sd.port, sd.w);
    } catch (InterruptedIOException e ) {
      return false; // retry timed out
    } catch (LogicalClockException e ) {
      e.printStackTrace();
      throw new ProtocolFailureException("Not valid clock used in session "+identity);
    }
    // Computes the compensated half roundtrip
    D = ((rc.t1() - rc.t0()) - (rc.tx() - rc.tr())) / 2;
    // reads the present precision, to detect emergency
    if ( ! lc.is_valid() )
    // in case the clock accuracy is unbound
      {
        mode = "INITIAL";
        psi = D - sd.min;
        t = (rc.tx() + sd.min + psi);
        delta_0 = psi + rc.precision();
        lc.set(t, delta_0,rc.read_baseline()); // set the logical clock!
      }
    else
    // in case the clock accuracy is bound
      {
        delta_sum = delta_0 + rc.precision(); // accuracy is the sum of local and remote
	alpha_max = rc.tr() - rc.t0() + delta_sum - sd.min;
	if ( alpha_max <= 0 ) 
	// this is an error condition
        {
	  System.err.println("Emergency: forward trip timing is inconsistent");
	  emergency = true;
	  errdelta = rc.t0() - rc.tr();
	}
	alpha_min = rc.tr() - rc.t0() - delta_sum - sd.min;
	if ( alpha_min < 0 )
          alpha_min = 0;
	beta_min = 2 * D - 2 * sd.min - alpha_max;
	if ( beta_min < 0 )
	  beta_min = 0;
	beta_max = 2 * D - 2 * sd.min - alpha_min;
	if ( beta_max <= 0 ) 
	// this is an error condition
	{
	  System.err.println("Emergency: backward trip timing is inconsistent");
	  emergency = true;
	  errdelta =  rc.tx() - rc.t1();
	}
	x = Math.abs (D - (sd.min + delta_sum));
	l1 = rc.t0() + D - x;
	l2 = rc.t0() + D + x;
	if ((rc.tr() < l1) || (rc.tr() > l2))
	{	  
	  mode = "ALA_CIU";
          if (alpha_max < beta_max)
  	    psi = alpha_max / 2;
	  else
            psi = beta_max / 2;
	}
	else
	{
	  mode = "CRISTIAN";
	  psi = D - sd.min;
	}
	if ( psi < 0 ) 
	// this is an error condition
	{
	  System.err.println("Emergency: new precision is inconsistent");
	  emergency = true;
	}
        t = (rc.tx() + sd.min + beta_min + psi);
        if ( delta_0 > psi + rc.precision() )
	{
	  delta_0 = psi + rc.precision();
	  lc.set(t, delta_0,rc.read_baseline()); // set the logical clock!
	}
	else
	  mode = "USELESS";
      }
    report.append("    Remote clock reading report (session " + identity + "):\n");
    report.append("    Local transmit time (t0): \t"  + new Date(rc.t0()) + " " + (rc.t0() % 1000) + " msecs\n");
    report.append("    Remote receive time (tr): \t"  + new Date(rc.tr()) + " " + (rc.tr() % 1000) + " msecs\n");
    report.append("    Remote transmit time (tx): \t" + new Date(rc.tx()) + " " + (rc.tx() % 1000) + " msecs\n");
    report.append("    Local receive time (t1): \t"   + new Date(rc.t1()) + " " + (rc.t1() % 1000) + " msecs\n");
    report.append("    Remote precision: \t"     + rc.precision() + " msecs\n");
    report.append("    Compensated roundtrip (2D): "+ 2*D + " msecs\n");
    report.append("    Retry report:\n");
    report.append("    Retry mode was " + mode + "\n");
    report.append("    New precision (delta_0) is\t" + delta_0 + " msecs\n");
    report.append("    New time (at t_1) is \t" + new Date(t)  + " " + (t % 1000) + " msecs\n");
    log.verbose(report.toString());
    if ( emergency ) 
      {
        report.append("    Emergency: restarting clock\n");
	throw new ProtocolFailureException("Session "+ identity + ": inconsistent timing");
      }
    return true;      
  }
}













